from unittest.mock import patch

import cohere.base_client
from cohere import RerankResponse, RerankResponseResultsItem

from autorag.nodes.passagereranker import cohere_reranker
from tests.autorag.nodes.passagereranker.test_passage_reranker_base import queries_example, contents_example, \
    scores_example, ids_example, base_reranker_test, project_dir, previous_result, base_reranker_node_test


async def mock_cohere_reranker(self, model, query, documents, top_n, **kwargs) -> RerankResponse:
    if query == queries_example[0]:
        return RerankResponse(
            results=[
                        RerankResponseResultsItem(index=1, relevance_score=0.8),
                        RerankResponseResultsItem(index=2, relevance_score=0.2),
                        RerankResponseResultsItem(index=0, relevance_score=0.1),
                    ][:top_n]
        )
    elif query == queries_example[1]:
        return RerankResponse(
            results=[
                        RerankResponseResultsItem(index=2, relevance_score=0.8),
                        RerankResponseResultsItem(index=0, relevance_score=0.2),
                        RerankResponseResultsItem(index=1, relevance_score=0.1),
                    ][:top_n]
        )
    response_items = [
        RerankResponseResultsItem(index=i, relevance_score=0.1 * i)
        for i in range(len(documents) - 1, -1, -1)
    ]
    return RerankResponse(results=response_items[:top_n])


@patch.object(cohere.base_client.AsyncBaseCohere, "rerank", mock_cohere_reranker)
def test_cohere_reranker():
    top_k = 3
    original_cohere_reranker = cohere_reranker.__wrapped__
    contents_result, id_result, score_result \
        = original_cohere_reranker(queries_example, contents_example, scores_example, ids_example, top_k)
    base_reranker_test(contents_result, id_result, score_result, top_k)


@patch.object(cohere.base_client.AsyncBaseCohere, "rerank", mock_cohere_reranker)
def test_cohere_reranker_batch_one():
    top_k = 3
    batch = 1
    original_cohere_reranker = cohere_reranker.__wrapped__
    contents_result, id_result, score_result \
        = original_cohere_reranker(queries_example, contents_example, scores_example, ids_example, top_k, batch=batch)
    base_reranker_test(contents_result, id_result, score_result, top_k)


@patch.object(cohere.base_client.AsyncBaseCohere, "rerank", mock_cohere_reranker)
def test_cohere_node():
    top_k = 1
    result_df = cohere_reranker(project_dir=project_dir, previous_result=previous_result, top_k=top_k)
    base_reranker_node_test(result_df, top_k)



import pytest
from autorag.schema.module import Module
from autorag.support import get_support_modules


# Test cases for supported module types
@pytest.mark.parametrize("module_type, expected_module", [
    ('bm25', get_support_modules('bm25')),
    ('vectordb', get_support_modules('vectordb')),
])
def test_module_from_dict_supported(module_type, expected_module):
    module_dict = {
        'module_type': module_type,
        'param1': 'value1',
        'param2': 'value2',
    }
    module = Module.from_dict(module_dict)
    assert module.module_type == module_type
    assert module.module == expected_module
    assert module.module_param == {k: v for k, v in module_dict.items() if k != 'module_type'}


# Test cases for unsupported module types
@pytest.mark.parametrize("module_type", [
    'unsupported_module',
    'another_unsupported_module',
])
def test_module_from_dict_unsupported(module_type):
    module_dict = {'module_type': module_type}
    with pytest.raises(KeyError):
        Module.from_dict(module_dict)



import tempfile

import pytest
from langchain_core.documents import Document

from autorag.data.corpus import langchain_documents_to_parquet
from tests.autorag.data.corpus.test_base import validate_corpus


@pytest.fixture
def parquet_filepath():
    with tempfile.NamedTemporaryFile(suffix='.parquet') as temp_file:
        yield temp_file.name


def test_langchain_documents_to_parquet(parquet_filepath):
    documents = [Document(page_content='test text', metadata={'key': 'value'}) for _ in range(5)]
    result_df = langchain_documents_to_parquet(documents, parquet_filepath, upsert=True)
    validate_corpus(result_df, 5, parquet_filepath)



import functools
import logging
from pathlib import Path
from typing import List, Union, Dict, Optional

import pandas as pd

from autorag.support import get_support_modules
from autorag.utils import result_to_dataframe, validate_qa_dataset

logger = logging.getLogger("AutoRAG")


def query_expansion_node(func):
    @functools.wraps(func)
    @result_to_dataframe(["queries"])
    def wrapper(
            project_dir: Union[str, Path],
            previous_result: pd.DataFrame,
            *args, **kwargs) -> List[List[str]]:
        logger.info(f"Running query expansion node - {func.__name__} module...")
        validate_qa_dataset(previous_result)

        # find queries columns
        assert "query" in previous_result.columns, "previous_result must have query column."
        queries = previous_result["query"].tolist()

        if func.__name__ == "pass_query_expansion":
            return func(queries=queries)

        # pop prompt from kwargs
        if "prompt" in kwargs.keys():
            prompt = kwargs.pop("prompt")
        else:
            prompt = ""

        # set generator module for query expansion
        generator_callable, generator_param = make_generator_callable_param(kwargs)

        # run query expansion function
        expanded_queries = func(queries=queries,
                                prompt=prompt,
                                generator_func=generator_callable,
                                generator_params=generator_param)
        return expanded_queries

    return wrapper


def make_generator_callable_param(generator_dict: Optional[Dict]):
    if 'generator_module_type' not in generator_dict.keys():
        generator_dict = {
            'generator_module_type': 'llama_index_llm',
            'llm': 'openai',
            'model': 'gpt-3.5-turbo',
        }
    module_str = generator_dict.pop('generator_module_type')
    module_callable = get_support_modules(module_str)
    module_param = generator_dict
    return module_callable, module_param



import asyncio
from unittest.mock import patch

from aioresponses import aioresponses

import autorag
from autorag.nodes.passagereranker import jina_reranker
from autorag.nodes.passagereranker.jina import jina_reranker_pure, JINA_API_URL
from tests.autorag.nodes.passagereranker.test_passage_reranker_base import queries_example, contents_example, \
    scores_example, ids_example, base_reranker_test, project_dir, previous_result, base_reranker_node_test


def test_jina_reranker_pure():
    with aioresponses() as m:
        mock_response = {
            "results": [
                {"index": 1, "relevance_score": 0.9},
                {"index": 0, "relevance_score": 0.2},
            ]
        }
        m.post(JINA_API_URL, payload=mock_response)
        content_result, id_result, score_result = asyncio.run(
            jina_reranker_pure(queries_example[0], contents_example[0], scores_example[0], ids_example[0], top_k=2,
                               api_key="mock_api_key"))
        assert len(content_result) == 2
        assert len(id_result) == 2
        assert len(score_result) == 2

        assert all([res in contents_example[0] for res in content_result])
        assert all([res in ids_example[0] for res in id_result])

        # check if the scores are sorted
        assert score_result[0] >= score_result[1]


async def mock_jina_reranker_pure(query, contents, scores, ids, top_k, api_key, **kwargs):
    if query == queries_example[0]:
        return [contents[1], contents[2], contents[0]][:top_k], [ids[1], ids[2], ids[0]][:top_k], [0.8, 0.2, 0.1][
                                                                                                  :top_k]
    elif query == queries_example[1]:
        return [contents[2], contents[0], contents[1]][:top_k], [ids[2], ids[0], ids[1]][:top_k], [0.8, 0.2, 0.1][
                                                                                                  :top_k]
    else:
        raise ValueError(f"Unexpected query: {query}")


@patch.object(autorag.nodes.passagereranker.jina, "jina_reranker_pure", mock_jina_reranker_pure)
def test_jina_reranker():
    top_k = 3
    original_jina_reranker = jina_reranker.__wrapped__
    contents_result, id_result, score_result \
        = original_jina_reranker(queries_example, contents_example, scores_example, ids_example, top_k)
    base_reranker_test(contents_result, id_result, score_result, top_k)


@patch.object(autorag.nodes.passagereranker.jina, "jina_reranker_pure", mock_jina_reranker_pure)
def test_jina_reranker_batch_one():
    top_k = 3
    batch = 1
    original_jina_reranker = jina_reranker.__wrapped__
    contents_result, id_result, score_result \
        = original_jina_reranker(queries_example, contents_example, scores_example, ids_example, top_k, batch=batch)
    base_reranker_test(contents_result, id_result, score_result, top_k)


@patch.object(autorag.nodes.passagereranker.jina, "jina_reranker_pure", mock_jina_reranker_pure)
def test_jina_reranker_node():
    top_k = 1
    result_df = jina_reranker(project_dir=project_dir, previous_result=previous_result, top_k=top_k)
    base_reranker_node_test(result_df, top_k)



import os

import nest_asyncio


def pytest_sessionstart(session):
    nest_asyncio.apply()
    os.environ["BM25"] = "bm25"
    os.environ["JINAAI_API_KEY"] = "mock_jinaai_api_key"
    os.environ["COHERE_API_KEY"] = "mock_cohere_api_key"



from .pass_passage_filter import pass_passage_filter
from .percentile_cutoff import percentile_cutoff
from .recency import recency_filter
from .similarity_percentile_cutoff import similarity_percentile_cutoff
from .similarity_threshold_cutoff import similarity_threshold_cutoff
from .threshold_cutoff import threshold_cutoff



import os
import shutil
import sys

import torch.cuda


def is_ubuntu() -> bool:
    try:
        with open("/etc/os-release") as f:
            for line in f:
                if line.startswith("NAME") and "Ubuntu" in line:
                    return True
    except FileNotFoundError:
        pass
    return False


def is_github_action() -> bool:
    return is_ubuntu() and not torch.cuda.is_available()


def main():
    paths = sys.path
    for path in paths:
        tests_dir = os.path.join(path, 'tests')
        if os.path.exists(tests_dir):
            print(f"Deleting {tests_dir}")
            shutil.rmtree(tests_dir)


if __name__ == "__main__":
    main()



import logging
from datetime import datetime, date
from typing import List, Tuple, Union

from autorag.nodes.passagefilter.base import passage_filter_node

logger = logging.getLogger("AutoRAG")


@passage_filter_node
def recency_filter(contents_list: List[List[str]],
                   scores_list: List[List[float]], ids_list: List[List[str]],
                   time_list: List[List[datetime]],
                   threshold: Union[datetime, date],
                   ) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Filter out the contents that are below the threshold datetime.
    If all contents are filtered, keep the only one recency content.
    If the threshold date format is incorrect, return the original contents.

    :param contents_list: The list of lists of contents to filter
    :param scores_list: The list of lists of scores retrieved
    :param ids_list: The list of lists of ids retrieved
    :param time_list: The list of lists of datetime retrieved
    :param threshold: The threshold to cut off
    :return: Tuple of lists containing the filtered contents, ids, and scores
    """
    if not (isinstance(threshold, datetime) or isinstance(threshold, date)):
        raise ValueError(f"Threshold should be a datetime object, but got {type(threshold)}")

    if not isinstance(threshold, datetime):
        threshold = datetime.combine(threshold, datetime.min.time())

    time_list = [
        list(map(lambda t: datetime.combine(t, datetime.min.time()) if not isinstance(t, datetime) else t, time))
        for time in time_list
    ]

    def sort_row(contents, scores, ids, time, _datetime_threshold):
        combined = list(zip(contents, scores, ids, time))
        combined_filtered = [item for item in combined if item[3] >= _datetime_threshold]

        if combined_filtered:
            remain_contents, remain_scores, remain_ids, _ = zip(*combined_filtered)
        else:
            combined.sort(key=lambda x: x[3], reverse=True)
            remain_contents, remain_scores, remain_ids, _ = zip(*combined[:1])

        return list(remain_contents), list(remain_ids), list(remain_scores)

    remain_contents_list, remain_ids_list, remain_scores_list = zip(
        *map(sort_row, contents_list, scores_list, ids_list, time_list, [threshold] * len(contents_list)))

    return remain_contents_list, remain_ids_list, remain_scores_list



from typing import List, Tuple

import numpy as np

from autorag.nodes.passagefilter.base import passage_filter_node


@passage_filter_node
def threshold_cutoff(queries: List[str], contents_list: List[List[str]],
                     scores_list: List[List[float]], ids_list: List[List[str]],
                     threshold: float, reverse: bool = False,
                     ) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Filters the contents, scores, and ids based on a previous result's score.
    Keeps at least one item per query if all scores are below the threshold.

    :param queries: List of query strings (not used in the current implementation).
    :param contents_list: List of content strings for each query.
    :param scores_list: List of scores for each content.
    :param ids_list: List of ids for each content.
    :param threshold: The minimum score to keep an item.
    :param reverse: If True, the lower the score, the better.
        Default is False.
    :return: Filtered lists of contents, ids, and scores.
    """
    remain_indices = list(map(lambda x: threshold_cutoff_pure(x, threshold, reverse), scores_list))

    remain_content_list = list(map(lambda c, idx: [c[i] for i in idx], contents_list, remain_indices))
    remain_scores_list = list(map(lambda s, idx: [s[i] for i in idx], scores_list, remain_indices))
    remain_ids_list = list(map(lambda _id, idx: [_id[i] for i in idx], ids_list, remain_indices))

    return remain_content_list, remain_ids_list, remain_scores_list


def threshold_cutoff_pure(scores_list: List[float],
                          threshold: float,
                          reverse: bool = False) -> List[int]:
    """
    Return indices that have to remain.
    Return at least one index if there is nothing to remain.

    :param scores_list: Each score
    :param threshold: The threshold to cut off
    :param reverse: If True, the lower the score, the better
        Default is False.
    :return: Indices to remain at the contents
    """
    if isinstance(scores_list, np.ndarray):
        if reverse:
            remain_indices = np.where(scores_list <= threshold)[0]
            default_index = np.argmin(scores_list)
        else:
            remain_indices = np.where(scores_list >= threshold)[0]
            default_index = np.argmax(scores_list)
    else:
        if reverse:
            remain_indices = [i for i, score in enumerate(scores_list) if score <= threshold]
            default_index = scores_list.index(min(scores_list))
        else:
            remain_indices = [i for i, score in enumerate(scores_list) if score >= threshold]
            default_index = scores_list.index(max(scores_list))

    return remain_indices.tolist() if isinstance(remain_indices, np.ndarray) and remain_indices.size > 0 else \
        remain_indices if remain_indices else [default_index]



from datetime import datetime
from typing import List, Tuple

from autorag.nodes.passagereranker.base import passage_reranker_node


@passage_reranker_node
def time_reranker(contents_list: List[List[str]],
                  scores_list: List[List[float]], ids_list: List[List[str]],
                  top_k: int, time_list: List[List[datetime]]
                  ) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank the passages based on merely the datetime of the passage.
    It uses 'last_modified_datetime' key in the corpus metadata,
    so the metadata should be in the format of {'last_modified_datetime': datetime.datetime} at the corpus data file.

    :param contents_list: The list of lists of contents
    :param scores_list: The list of lists of scores from the initial ranking
    :param ids_list: The list of lists of ids
    :param top_k: The number of passages to be retrieved after reranking
    :param time_list: The metadata list of lists of datetime.datetime
        It automatically extracts the 'last_modified_datetime' key from the metadata in the corpus data.
    :return: The reranked contents, ids, and scores
    """
    def sort_row(contents, scores, ids, time, top_k):
        combined = list(zip(contents, scores, ids, time))
        combined.sort(key=lambda x: x[3], reverse=True)
        sorted_contents, sorted_scores, sorted_ids, _ = zip(*combined)
        return list(sorted_contents)[:top_k], list(sorted_scores)[:top_k], list(sorted_ids)[:top_k]

    reranked_contents, reranked_scores, reranked_ids = zip(
        *map(sort_row, contents_list, scores_list, ids_list, time_list, [top_k] * len(contents_list)))

    return list(reranked_contents), list(reranked_ids), list(reranked_scores)



from autorag.nodes.passagereranker import time_reranker

from tests.autorag.nodes.passagereranker.test_passage_reranker_base import contents_example, \
    scores_example, ids_example, base_reranker_test, previous_result, base_reranker_node_test, time_list, \
    project_dir_with_corpus


def test_time_reranker():
    top_k = 2

    original_time_reranker = time_reranker.__wrapped__
    contents_result, id_result, score_result \
        = original_time_reranker(contents_example, scores_example, ids_example, top_k, time_list)
    assert id_result[0] == [ids_example[0][1], ids_example[0][2]]
    assert id_result[1] == [ids_example[1][2], ids_example[1][1]]
    assert contents_result[0] == [contents_example[0][1], contents_example[0][2]]
    assert contents_result[1] == [contents_example[1][2], contents_example[1][1]]
    assert score_result[0] == [0.8, 0.1]
    assert score_result[1] == [0.7, 0.2]
    base_reranker_test(contents_result, id_result, score_result, top_k)


def test_time_reranker_node(project_dir_with_corpus):
    top_k = 1
    result_df = time_reranker(project_dir=project_dir_with_corpus, previous_result=previous_result, top_k=top_k)
    base_reranker_node_test(result_df, top_k)



from autorag.nodes.queryexpansion import pass_query_expansion
from tests.autorag.nodes.queryexpansion.test_query_expansion_base import project_dir, previous_result


def test_pass_query_expansion():
    result_df = pass_query_expansion(project_dir=project_dir, previous_result=previous_result)
    assert len(result_df) == 5
    queries = result_df["queries"].tolist()
    assert all(isinstance(query, list) for query in queries)
    assert len(queries[0]) == 1
    assert queries == list(map(lambda x: [x], previous_result['query'].tolist()))



from .cohere import cohere_reranker
from .colbert import colbert_reranker
from .flag_embedding import flag_embedding_reranker
from .flag_embedding_llm import flag_embedding_llm_reranker
from .jina import jina_reranker
from .koreranker import koreranker
from .monot5 import monot5
from .pass_reranker import pass_reranker
from .rankgpt import rankgpt
from .sentence_transformer import sentence_transformer_reranker
from .tart.tart import tart
from .time_reranker import time_reranker
from .upr import upr



import pandas as pd
import pytest

from autorag.nodes.generator import vllm
from tests.autorag.nodes.generator.test_generator_base import (prompts, check_generated_texts, check_generated_tokens,
                                                               check_generated_log_probs)


@pytest.mark.skip(reason="This test needs CUDA enabled machine with vllm installed.")
def test_vllm():
    previous_result = pd.DataFrame(
        {
            'prompts': prompts,
            'qid': ['id-1', 'id-2', 'id-3']
        })
    result_df = vllm(
        project_dir='.', previous_result=previous_result,
        llm='facebook/opt-125m', max_tokens=5, temperature=0.5,
    )
    tokens = result_df['generated_tokens'].tolist()
    log_probs = result_df['generated_log_probs'].tolist()
    check_generated_texts(result_df['generated_texts'].tolist())
    check_generated_tokens(tokens)
    check_generated_log_probs(log_probs)
    assert all(len(tokens[i]) == len(log_probs[i]) for i in range(len(tokens)))



from typing import List, Dict, Callable

import pandas as pd

from autorag.nodes.queryexpansion.base import query_expansion_node

hyde_prompt = "Please write a passage to answer the question"


@query_expansion_node
def hyde(queries: List[str],
         generator_func: Callable,
         generator_params: Dict,
         prompt: str = hyde_prompt) -> List[List[str]]:
    """
    HyDE, which inspired by "Precise Zero-shot Dense Retrieval without Relevance Labels" (https://arxiv.org/pdf/2212.10496.pdf)
    LLM model creates a hypothetical passage.
    And then, retrieve passages using hypothetical passage as a query.
    :param queries: List[str], queries to retrieve.
    :param generator_func: Callable, generator functions.
    :param generator_params: Dict, generator parameters.
    :param prompt: prompt to use when generating hypothetical passage
    :return: List[List[str]], List of hyde results.
    """
    full_prompts = list(
        map(lambda x: (prompt if not bool(prompt) else hyde_prompt) + f"\nQuestion: {x}\nPassage:", queries))
    input_df = pd.DataFrame({"prompts": full_prompts})
    result_df = generator_func(project_dir=None, previous_result=input_df, **generator_params)
    answers = result_df['generated_texts'].tolist()
    results = list(map(lambda x: [x], answers))
    return results



import inspect
from copy import deepcopy
from typing import List, Tuple

import torch

from autorag.nodes.generator.base import generator_node


@generator_node
def vllm(prompts: List[str], llm: str, **kwargs) -> Tuple[List[str], List[List[int]], List[List[float]]]:
    """
    Vllm module.
    It gets the VLLM instance, and returns generated texts by the input prompt.
    You can set logprobs to get the log probs of the generated text.
    Default logprobs is 1.

    :param prompts: A list of prompts.
    :param llm: Model name of vLLM.
    :param kwargs: The extra parameters for generating the text.
    :return: A tuple of three elements.
        The first element is a list of generated text.
        The second element is a list of generated text's token ids.
        The third element is a list of generated text's log probs.
    """
    try:
        from vllm.outputs import RequestOutput
        from vllm.sequence import SampleLogprobs
        from vllm import LLM, SamplingParams
    except ImportError:
        raise ImportError("Please install vllm library. You can install it by running `pip install vllm`.")

    input_kwargs = deepcopy(kwargs)
    vllm_model = make_vllm_instance(llm, input_kwargs)

    if 'logprobs' not in input_kwargs:
        input_kwargs['logprobs'] = 1

    generate_params = SamplingParams(**input_kwargs)
    results: List[RequestOutput] = vllm_model.generate(prompts, generate_params)
    generated_texts = list(map(lambda x: x.outputs[0].text, results))
    generated_token_ids = list(map(lambda x: x.outputs[0].token_ids, results))
    log_probs: List[SampleLogprobs] = list(map(lambda x: x.outputs[0].logprobs, results))
    generated_log_probs = list(map(lambda x: list(map(
        lambda y: y[0][y[1]].logprob, zip(x[0], x[1])
    )), zip(log_probs, generated_token_ids)))
    destroy_vllm_instance(vllm_model)
    return generated_texts, generated_token_ids, generated_log_probs


def make_vllm_instance(llm: str, input_args):
    from vllm import LLM
    model_from_args = input_args.pop('model', None)
    model = llm if model_from_args is None else model_from_args
    init_params = inspect.signature(LLM.__init__).parameters.values()
    keyword_init_params = [param.name for param in init_params if param.kind == param.KEYWORD_ONLY]
    input_kwargs = {}
    for param in keyword_init_params:
        v = input_args.pop(param, None)
        if v is not None:
            input_kwargs[param] = v
    return LLM(model, **input_kwargs)


def destroy_vllm_instance(vllm_instance):
    if torch.cuda.is_available():
        from vllm.distributed.parallel_state import (
            destroy_model_parallel,
        )

        destroy_model_parallel()
        del vllm_instance
        torch.cuda.synchronize()
        torch.cuda.empty_cache()
    else:
        del vllm_instance



import asyncio
from typing import List, Tuple

from llama_index.core.service_context_elements.llm_predictor import LLMPredictorType
from transformers import AutoTokenizer

from autorag.nodes.generator.base import generator_node
from autorag.utils.util import process_batch


@generator_node
def llama_index_llm(prompts: List[str], llm: LLMPredictorType, batch: int = 16) -> Tuple[List[str], List[List[int]], List[List[float]]]:
    """
    Llama Index LLM module.
    It gets the LLM instance from llama index, and returns generated text by the input prompt.
    It does not generate the right log probs, but it returns the pseudo log probs,
    which is not meant to be used for other modules.

    :param prompts: A list of prompts.
    :param llm: A llama index LLM instance.
    :param batch: The batch size for llm.
        Set low if you face some errors.
    :return: A tuple of three elements.
        The first element is a list of generated text.
        The second element is a list of generated text's token ids, used tokenizer is GPT2Tokenizer.
        The third element is a list of generated text's pseudo log probs.
    """
    tasks = [llm.acomplete(prompt) for prompt in prompts]
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(process_batch(tasks, batch_size=batch))

    generated_texts = list(map(lambda x: x.text, results))
    tokenizer = AutoTokenizer.from_pretrained("gpt2", use_fast=False)
    tokenized_ids = tokenizer(generated_texts).data['input_ids']
    pseudo_log_probs = list(map(lambda x: [0.5] * len(x), tokenized_ids))
    return generated_texts, tokenized_ids, pseudo_log_probs



import functools
import itertools
import logging
import os
from pathlib import Path
from typing import List, Union, Tuple

import numpy as np
import pandas as pd
import torch

from autorag import embedding_models
from autorag.evaluate.metric.util import calculate_cosine_similarity
from autorag.utils import (result_to_dataframe, validate_qa_dataset, fetch_contents, sort_by_scores,
                           validate_corpus_dataset, cast_corpus_dataset)
from autorag.utils.util import reconstruct_list, filter_dict_keys, select_top_k

logger = logging.getLogger("AutoRAG")


def passage_augmenter_node(func):
    @functools.wraps(func)
    @result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
    def wrapper(
            project_dir: Union[str, Path],
            previous_result: pd.DataFrame,
            *args, **kwargs) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
        logger.info(f"Running passage augmenter node - {func.__name__} module...")
        validate_qa_dataset(previous_result)
        data_dir = os.path.join(project_dir, "data")

        # find queries columns
        assert "query" in previous_result.columns, "previous_result must have query column."
        queries = previous_result["query"].tolist()

        # find ids columns
        assert "retrieved_ids" in previous_result.columns, "previous_result must have retrieved_ids column."
        ids = previous_result["retrieved_ids"].tolist()

        corpus_df = pd.read_parquet(os.path.join(data_dir, "corpus.parquet"))
        validate_corpus_dataset(corpus_df)

        # get top_k
        top_k = kwargs.pop("top_k")

        if func.__name__ == 'prev_next_augmenter':
            corpus_df = cast_corpus_dataset(corpus_df)
            slim_corpus_df = corpus_df[["doc_id", "metadata"]]
            slim_corpus_df['metadata'] = slim_corpus_df['metadata'].apply(filter_dict_keys, keys=['prev_id', 'next_id'])

            mode = kwargs.pop("mode", 'both')
            num_passages = kwargs.pop("num_passages", 1)

            # get augmented ids
            augmented_ids = func(ids_list=ids, corpus_df=slim_corpus_df, mode=mode, num_passages=num_passages)

            # fetch contents from corpus to use augmented ids
            augmented_contents = fetch_contents(corpus_df, augmented_ids)

            # set embedding model for getting scores
            embedding_model_str = kwargs.pop("embedding_model", 'openai')
            query_embeddings, contents_embeddings = embedding_query_content(queries, augmented_contents,
                                                                            embedding_model_str, batch=128)

            # get scores from calculated cosine similarity
            augmented_scores = [
                np.array([calculate_cosine_similarity(query_embedding, x) for x in content_embeddings]).tolist()
                for query_embedding, content_embeddings in zip(query_embeddings, contents_embeddings)]
        else:
            contents = previous_result["retrieved_contents"].tolist()
            scores = previous_result["retrieve_scores"].tolist()

            augmented_ids, augmented_contents, augmented_scores = func(ids_list=ids, contents_list=contents,
                                                                       scores_list=scores, **kwargs)

        # sort by scores
        df = pd.DataFrame({
            'contents': augmented_contents,
            'ids': augmented_ids,
            'scores': augmented_scores,
        })
        df[['contents', 'ids', 'scores']] = df.apply(sort_by_scores, axis=1, result_type='expand')

        # select by top_k
        results = select_top_k(df, ['contents', 'ids', 'scores'], top_k)

        return results['contents'].tolist(), results['ids'].tolist(), results['scores'].tolist()

    return wrapper


def embedding_query_content(queries: List[str], contents_list: List[List[str]],
                            embedding_model: str, batch: int = 128):
    embedding_model = embedding_models[embedding_model]

    # Embedding using batch
    embedding_model.embed_batch_size = batch
    query_embeddings = embedding_model.get_text_embedding_batch(queries)

    content_lengths = list(map(len, contents_list))
    content_embeddings_flatten = embedding_model.get_text_embedding_batch(list(
        itertools.chain.from_iterable(contents_list)))
    content_embeddings = reconstruct_list(content_embeddings_flatten, content_lengths)

    del embedding_model
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
    return query_embeddings, content_embeddings



import os
import pathlib

import pandas as pd

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent.parent.parent
project_dir = os.path.join(root_dir, "resources", "sample_project")
qa_data = pd.read_parquet(os.path.join(project_dir, "data", "qa.parquet"))
corpus_data = pd.read_parquet(os.path.join(project_dir, "data", "corpus.parquet"))
doc_id_list = corpus_data["doc_id"].tolist()
previous_result = pd.DataFrame({
    'qid': [1, 2],
    'query': ['What is the capital of France?', 'How many members are in Newjeans?'],
    'retrieved_ids': [[doc_id_list[1], doc_id_list[3]],
                      [doc_id_list[5], doc_id_list[7]]],
    'retrieved_contents': [['Paris is the capital of France.', 'Paris is one of the capital from France. Isn\'t it?'],
                           ['Newjeans has 5 members.', 'Danielle is one of the members of Newjeans.']],
    'retrieve_scores': [[0.8, 0.1], [0.2, 0.1]],
    'retrieval_gt': [1, 1],
    'generation_gt': ['answer', 'answer'],
    'retrieval_f1': [0.4, 1.0],
    'retrieval_recall': [1.0, 0.3]
})
ids_list = [[doc_id_list[1]], [doc_id_list[3]], [doc_id_list[0], doc_id_list[29]]]



import functools
import logging
from pathlib import Path
from typing import List, Union

import pandas as pd

from autorag.utils import result_to_dataframe

logger = logging.getLogger("AutoRAG")


def prompt_maker_node(func):
    @functools.wraps(func)
    @result_to_dataframe(["prompts"])
    def wrapper(
            project_dir: Union[str, Path],
            previous_result: pd.DataFrame,
            *args, **kwargs) -> List[str]:
        logger.info(f"Running prompt maker node - {func.__name__} module...")
        # get query and retrieved contents from previous_result
        assert "query" in previous_result.columns, "previous_result must have query column."
        assert "retrieved_contents" in previous_result.columns, "previous_result must have retrieved_contents column."
        query = previous_result["query"].tolist()
        retrieved_contents = previous_result["retrieved_contents"].tolist()
        prompt = kwargs.pop("prompt")

        if func.__name__ == 'fstring':
            return func(prompt, query, retrieved_contents)
        elif func.__name__ == 'long_context_reorder':
            assert "retrieve_scores" in previous_result.columns, "previous_result must have retrieve_scores column."
            retrieve_scores = previous_result["retrieve_scores"].tolist()
            return func(prompt, query, retrieved_contents, retrieve_scores)
        else:
            raise NotImplementedError(f"Module {func.__name__} is not implemented or not supported.")

    return wrapper



import logging
import os
import uuid
from copy import deepcopy
from typing import Optional, Dict, List

import pandas as pd
import uvicorn
import yaml
from fastapi import FastAPI
from pydantic import BaseModel

from autorag.support import get_support_modules
from autorag.utils.util import load_summary_file

logger = logging.getLogger("AutoRAG")


def extract_node_line_names(config_dict: Dict) -> List[str]:
    """
    Extract node line names with the given config dictionary order.

    :param config_dict: The yaml configuration dict for the pipeline.
        You can load this to access trail_folder/config.yaml.
    :return: The list of node line names.
        It is the order of the node line names in the pipeline.
    """
    return [node_line['node_line_name'] for node_line in config_dict['node_lines']]


def extract_node_strategy(config_dict: Dict) -> Dict:
    """
    Extract node strategies with the given config dictionary.
    The return value is a dictionary of node type and its strategy.

    :param config_dict: The yaml configuration dict for the pipeline.
        You can load this to access trail_folder/config.yaml.
    :return: Key is node_type and value is strategy dict.
    """
    return {node['node_type']: node.get('strategy', {})
            for node_line in config_dict['node_lines']
            for node in node_line['nodes']}


def summary_df_to_yaml(summary_df: pd.DataFrame, config_dict: Dict) -> Dict:
    """
    Convert trial summary dataframe to config yaml file.

    :param summary_df: The trial summary dataframe of the evaluated trial.
    :param config_dict: The yaml configuration dict for the pipeline.
        You can load this to access trail_folder/config.yaml.
    :return: Dictionary of config yaml file.
        You can save this dictionary to yaml file.
    """

    # summary_df columns : 'node_line_name', 'node_type', 'best_module_filename',
    #                      'best_module_name', 'best_module_params', 'best_execution_time'
    node_line_names = extract_node_line_names(config_dict)
    node_strategies = extract_node_strategy(config_dict)
    strategy_df = pd.DataFrame({
        'node_type': list(node_strategies.keys()),
        'strategy': list(node_strategies.values())
    })
    summary_df = summary_df.merge(strategy_df, on='node_type', how='left')
    summary_df['categorical_node_line_name'] = pd.Categorical(summary_df['node_line_name'], categories=node_line_names,
                                                              ordered=True)
    summary_df = summary_df.sort_values(by='categorical_node_line_name')
    grouped = summary_df.groupby('categorical_node_line_name')

    node_lines = [
        {
            'node_line_name': node_line_name,
            'nodes': [
                {
                    'node_type': row['node_type'],
                    'strategy': row['strategy'],
                    'modules': [{
                        'module_type': row['best_module_name'],
                        **row['best_module_params']
                    }]
                }
                for _, row in node_line.iterrows()
            ]
        }
        for node_line_name, node_line in grouped
    ]
    return {'node_lines': node_lines}


def extract_best_config(trial_path: str, output_path: Optional[str] = None) -> Dict:
    """
    Extract the optimal pipeline from evaluated trial.

    :param trial_path: The path to the trial directory that you want to extract the pipeline from.
        Must already be evaluated.
    :param output_path: Output path that pipeline yaml file will be saved.
        Must be .yaml or .yml file.
        If None, it does not save yaml file and just return dict values.
        Default is None.
    :return: The dictionary of the extracted pipeline.
    """
    summary_path = os.path.join(trial_path, 'summary.csv')
    if not os.path.exists(summary_path):
        raise ValueError(f"summary.csv does not exist in {trial_path}.")
    trial_summary_df = load_summary_file(summary_path, dict_columns=['best_module_params'])
    config_yaml_path = os.path.join(trial_path, 'config.yaml')
    with open(config_yaml_path, 'r') as f:
        config_dict = yaml.safe_load(f)
    yaml_dict = summary_df_to_yaml(trial_summary_df, config_dict)
    if output_path is not None:
        with open(output_path, 'w') as f:
            yaml.dump(yaml_dict, f)
    return yaml_dict


class Runner:
    def __init__(self, config: Dict, project_dir: Optional[str] = None):
        self.config = config
        self.project_dir = os.getcwd() if project_dir is None else project_dir
        self.app = FastAPI()
        self.__add_api_route()

    @classmethod
    def from_yaml(cls, yaml_path: str, project_dir: Optional[str] = None):
        """
        Load Runner from yaml file.
        Must be extracted yaml file from evaluated trial using extract_best_config method.

        :param yaml_path: The path of the yaml file.
        :param project_dir: The path of the project directory.
            Default is the current directory.
        :return: Initialized Runner.
        """
        with open(yaml_path, 'r') as f:
            try:
                config = yaml.safe_load(f)
            except yaml.YAMLError as exc:
                logger.error(exc)
                raise exc
        return cls(config, project_dir=project_dir)

    @classmethod
    def from_trial_folder(cls, trial_path: str):
        """
        Load Runner from evaluated trial folder.
        Must already be evaluated using Evaluator class.
        It sets the project_dir as the parent directory of the trial folder.

        :param trial_path: The path of the trial folder.
        :return: Initialized Runner.
        """
        config = extract_best_config(trial_path)
        return cls(config, project_dir=os.path.dirname(trial_path))

    def run(self, query: str, result_column: str = "generated_texts"):
        """
        Run the pipeline with query.
        The loaded pipeline must start with a single query,
        so the first module of the pipeline must be `query_expansion` or `retrieval` module.

        :param query: The query of the user.
        :param result_column: The result column name for the answer.
            Default is `generated_texts`, which is the output of the `generation` module.
        :return: The result of the pipeline.
        """
        node_lines = deepcopy(self.config['node_lines'])
        previous_result = pd.DataFrame({
            'qid': str(uuid.uuid4()),
            'query': [query],
            'retrieval_gt': [[]],
            'generation_gt': [''],
        })  # pseudo qa data for execution
        for node_line in node_lines:
            for node in node_line['nodes']:
                if len(node['modules']) != 1:
                    raise ValueError("The number of modules in a node must be 1 for using runner."
                                     "Please use extract_best_config method for extracting yaml file from evaluated trial.")
                module = node['modules'][0]
                module_type = module.pop('module_type')
                module_params = module
                new_result = get_support_modules(module_type)(
                    project_dir=self.project_dir,
                    previous_result=previous_result,
                    **module_params
                )
                duplicated_columns = previous_result.columns.intersection(new_result.columns)
                drop_previous_result = previous_result.drop(columns=duplicated_columns)
                previous_result = pd.concat([drop_previous_result, new_result], axis=1)

        return previous_result[result_column].tolist()[0]

    def __add_api_route(self):

        @self.app.post("/run")
        async def run_pipeline(runner_input: RunnerInput):
            query = runner_input.query
            result_column = runner_input.result_column
            result = self.run(query, result_column)
            return {result_column: result}

    def run_api_server(self, host: str = '0.0.0.0', port: int = 8000, **kwargs):
        """
        Run the pipeline as api server.
        You can send POST request to `http://host:port/run` with json body like below:

        .. Code:: json

            {
                "Query": "your query",
                "result_column": "answer"
            }

        And it returns json response like below:

        .. Code:: json

            {
                "answer": "your answer"
            }

        :param host: The host of the api server.
        :param port: The port of the api server.
        :param kwargs: Other arguments for uvicorn.run.
        """
        logger.info(f"Run api server at {host}:{port}")
        uvicorn.run(self.app, host=host, port=port, **kwargs)


class RunnerInput(BaseModel):
    query: str
    result_column: str = "answer"



from copy import deepcopy
from dataclasses import dataclass, field
from typing import Callable, Dict

from autorag.support import get_support_modules


@dataclass
class Module:
    module_type: str
    module_param: Dict
    module: Callable = field(init=False)

    def __post_init__(self):
        self.module = get_support_modules(self.module_type)
        if self.module is None:
            raise ValueError(f"Module type {self.module_type} is not supported.")

    @classmethod
    def from_dict(cls, module_dict: Dict) -> 'Module':
        _module_dict = deepcopy(module_dict)
        module_type = _module_dict.pop('module_type')
        module_params = _module_dict
        return cls(module_type, module_params)



import logging
import os
from typing import List, Tuple, Dict

import numpy as np
import pandas as pd
import ray
import torch
from transformers import T5Tokenizer, T5ForConditionalGeneration

from autorag.nodes.passagereranker.base import passage_reranker_node
from autorag.utils.util import select_top_k, sort_by_scores

logger = logging.getLogger("AutoRAG")


@passage_reranker_node
def upr(queries: List[str], contents_list: List[List[str]],
        scores_list: List[List[float]], ids_list: List[List[str]],
        top_k: int, use_bf16: bool = False,
        prefix_prompt: str = "Passage: ",
        suffix_prompt: str = "Please write a question based on this passage.",
        num_gpus: int = 1) \
        -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents based on their relevance to a query using UPR.
    UPR is a reranker based on UPR (https://github.com/DevSinghSachan/unsupervised-passage-reranking).
    The language model will make a question based on the passage and rerank the passages by the likelihood of the question.
    The default model is t5-large.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param use_bf16: Whether to use bfloat16 for the model. Default is False.
    :param prefix_prompt: The prefix prompt for the language model that generates question for reranking.
        Default is "Passage: ".
        The prefix prompt serves as the initial context or instruction for the language model.
        It sets the stage for what is expected in the output
    :param suffix_prompt: The suffix prompt for the language model that generates question for reranking.
        Default is "Please write a question based on this passage.".
        The suffix prompt provides a cue or a closing instruction to the language model,
            signaling how to conclude the generated text or what format to follow at the end.
    :param num_gpus: The number of GPUs that you want to use. Default is 1.
        Set to 1 if you use a CPU or a single GPU.
    :return: tuple of lists containing the reranked contents, ids, and scores
    """
    df = pd.DataFrame({
        'query': queries,
        'contents': contents_list,
        'ids': ids_list,
    })
    ds = ray.data.from_pandas(df)

    scorer = UPRScorer(suffix_prompt=suffix_prompt, prefix_prompt=prefix_prompt, use_bf16=use_bf16)

    if torch.cuda.is_available():
        score_batch = ds.map_batches(scorer, batch_size=1,
                                     concurrency=num_gpus, num_gpus=1)
    else:
        score_batch = ds.map_batches(scorer,
                                     batch_size=1,
                                     concurrency=min(len(df), os.cpu_count()), num_cpus=os.cpu_count())
    scores = score_batch.to_pandas()['output'].tolist()  # converted to a flatten list of scores

    del scorer

    explode_df = df.explode('contents')
    explode_df['scores'] = scores
    df['scores'] = explode_df.groupby(level=0, sort=False)['scores'].apply(list).tolist()
    df[['contents', 'ids', 'scores']] = df.apply(lambda x: sort_by_scores(x, reverse=False), axis=1,
                                                 result_type='expand')
    results = select_top_k(df, ['contents', 'ids', 'scores'], top_k)
    return results['contents'].tolist(), results['ids'].tolist(), results['scores'].tolist()


class UPRScorer:
    def __init__(self, suffix_prompt: str, prefix_prompt: str,
                 use_bf16: bool = False):
        model_name = "t5-large"
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.tokenizer = T5Tokenizer.from_pretrained(model_name)
        self.model = T5ForConditionalGeneration.from_pretrained(model_name,
                                                                torch_dtype=torch.bfloat16
                                                                if use_bf16 else torch.float32).to(self.device)
        self.suffix_prompt = suffix_prompt
        self.prefix_prompt = prefix_prompt

    def __call__(self, batch: Dict[str, np.ndarray]):
        query_token = self.tokenizer(batch['query'][0], max_length=128, truncation=True, return_tensors='pt')
        contents = batch['contents'][0]
        prompts = list(map(lambda content: f"{self.prefix_prompt} {content} {self.suffix_prompt}", contents))
        prompt_token_outputs = self.tokenizer(prompts, padding='longest',
                                              max_length=512,
                                              pad_to_multiple_of=8,
                                              truncation=True,
                                              return_tensors='pt')

        query_input_ids = torch.repeat_interleave(query_token['input_ids'], len(contents),
                                                  dim=0).to(self.device)

        with torch.no_grad():
            logits = self.model(input_ids=prompt_token_outputs['input_ids'].to(self.device),
                                attention_mask=prompt_token_outputs['attention_mask'].to(self.device),
                                labels=query_input_ids).logits
        log_softmax = torch.nn.functional.log_softmax(logits, dim=-1)
        nll = -log_softmax.gather(2, query_input_ids.unsqueeze(2)).squeeze(2)
        avg_nll = torch.sum(nll, dim=1)
        return {"output": avg_nll.tolist()}

    def __del__(self):
        del self.model
        del self.tokenizer
        if torch.cuda.is_available():
            torch.cuda.empty_cache()



import uuid
from datetime import datetime
from typing import List, Optional, Dict

import pandas as pd
from llama_index.core import Document
from llama_index.core.schema import TextNode, NodeRelationship

from autorag.data.utils.util import add_essential_metadata
from autorag.utils.util import save_parquet_safe


def llama_documents_to_parquet(llama_documents: List[Document],
                               output_filepath: Optional[str] = None,
                               upsert: bool = False) -> pd.DataFrame:
    """
    Llama Index documents to corpus dataframe.
    Corpus dataframe will be saved to filepath(file_dir/filename) if given.
    Return corpus dataframe whether the filepath is given.
    You can use this method to create corpus.parquet after load and chunk using Llama Index.

    :param llama_documents: List[Document]
    :param output_filepath: Optional filepath to save the parquet file.
        If None, the function will return the processed_data as pd.DataFrame, but do not save as parquet.
        File directory must exist. File extension must be .parquet
    :param upsert: If true, the function will overwrite the existing file if it exists.
        Default is False.
    :return: Corpus data as pd.DataFrame
    """

    doc_lst = pd.DataFrame(list(map(lambda doc: {
        'doc_id': str(uuid.uuid4()),
        'contents': doc.text,
        'metadata': add_essential_metadata(doc.metadata)
    }, llama_documents)))

    processed_df = pd.DataFrame(doc_lst)

    if output_filepath is not None:
        save_parquet_safe(processed_df, output_filepath, upsert=upsert)

    return processed_df


def llama_text_node_to_parquet(text_nodes: List[TextNode],
                               output_filepath: Optional[str] = None,
                               upsert: bool = False) -> pd.DataFrame:
    """
    Llama Index text nodes to corpus dataframe.
    Corpus dataframe will be saved to filepath(file_dir/filename) if given.
    Return corpus dataframe whether the filepath is given.
    You can use this method to create corpus.parquet after load and chunk using Llama Index.

    :param text_nodes: List of llama index text nodes.
    :param output_filepath: Optional filepath to save the parquet file.
        If None, the function will return the processed_data as pd.DataFrame, but do not save as parquet.
        File directory must exist. File extension must be .parquet
    :param upsert: If true, the function will overwrite the existing file if it exists.
        Default is False.
    :return: Corpus data as pd.DataFrame
    """
    corpus_df = pd.DataFrame(list(map(lambda node: {
        'doc_id': node.node_id,
        'contents': node.text,
        'metadata': add_essential_metadata_llama_text_node(node.metadata, node.relationships)
    }, text_nodes)))

    if output_filepath is not None:
        save_parquet_safe(corpus_df, output_filepath, upsert=upsert)

    return corpus_df


def add_essential_metadata_llama_text_node(metadata: Dict, relationships: Dict) -> Dict:
    if 'last_modified_datetime' not in metadata:
        metadata['last_modified_datetime'] = datetime.now()

    if 'prev_id' not in metadata:
        if NodeRelationship.PREVIOUS in relationships:
            prev_node = relationships.get(NodeRelationship.PREVIOUS, None)
            if prev_node:
                metadata['prev_id'] = prev_node.node_id

    if 'next_id' not in metadata:
        if NodeRelationship.NEXT in relationships:
            next_node = relationships.get(NodeRelationship.NEXT, None)
            if next_node:
                metadata['next_id'] = next_node.node_id
    return metadata



from typing import Tuple, List

import pandas as pd

from autorag.nodes.retrieval import retrieval_node


@retrieval_node
def hybrid_cc(
        ids: Tuple,
        scores: Tuple,
        top_k: int,
        weights: Tuple = (0.5, 0.5)) -> Tuple[List[List[str]], List[List[float]]]:
    """
    Hybrid CC function.
    CC (convex combination) is a method to fuse multiple retrieval results.
    It is a method that first normalizes the scores of each retrieval result,
    and then combines them with the given weights.
    To use this function, you must input ids and scores as tuple.
    It is uniquer than other retrieval modules, because it does not really execute retrieval,
    but just fuse the results of other retrieval functions.
    So you have to run more than two retrieval modules before running this function.
    And collect ids and scores result from each retrieval module.
    Make it as tuple and input it to this function.

    :param ids: The tuple of ids that you want to fuse.
        The length of this must be the same as the length of scores.
    :param scores: The retrieve scores that you want to fuse.
        The length of this must be the same as the length of ids.
    :param top_k: The number of passages to be retrieved.
    :param weights: Weight for each retrieval result.
        Default is (0.5, 0.5).
        You must set its length as the same as the length of ids and scores.
        Plus, the sum of the weights must be 1.
    :return: The tuple of ids and fused scores that fused by CC.
    """
    assert len(ids) == len(scores), "The length of ids and scores must be the same."
    assert len(ids) == len(weights), "The length of weights must be the same as the length of ids."
    assert len(ids) > 1, "You must input more than one retrieval results."
    assert top_k > 0, "top_k must be greater than 0."
    assert sum(weights) == 1, "The sum of weights must be 1."

    id_df = pd.DataFrame({f'id_{i}': id_list for i, id_list in enumerate(ids)})
    score_df = pd.DataFrame({f'score_{i}': score_list for i, score_list in enumerate(scores)})
    df = pd.concat([id_df, score_df], axis=1)

    def cc_pure_apply(row):
        ids_tuple = tuple(row[[f'id_{i}' for i in range(len(ids))]].values)
        scores_tuple = tuple(row[[f'score_{i}' for i in range(len(scores))]].values)
        return pd.Series(cc_pure(ids_tuple, scores_tuple, weights, top_k))

    df[['cc_id', 'cc_score']] = df.apply(cc_pure_apply, axis=1)
    return df['cc_id'].tolist(), df['cc_score'].tolist()


def cc_pure(ids: Tuple, scores: Tuple, weights: Tuple, top_k: int) -> Tuple[
    List[str], List[float]]:
    df = pd.concat([pd.Series(dict(zip(_id, score))) for _id, score in zip(ids, scores)], axis=1)
    normalized_scores = (df - df.min()) / (df.max() - df.min())
    normalized_scores = normalized_scores.fillna(0)
    normalized_scores['weighted_sum'] = normalized_scores.mul(weights).sum(axis=1)
    normalized_scores = normalized_scores.sort_values(by='weighted_sum', ascending=False)
    return normalized_scores.index.tolist()[:top_k], normalized_scores['weighted_sum'][:top_k].tolist()



import uuid
from typing import List, Optional

import pandas as pd
from langchain_core.documents import Document

from autorag.data.utils.util import add_essential_metadata
from autorag.utils.util import save_parquet_safe


def langchain_documents_to_parquet(langchain_documents: List[Document],
                                   output_filepath: Optional[str] = None,
                                   upsert: bool = False) -> pd.DataFrame:
    """
    Langchain documents to corpus dataframe.
    Corpus dataframe will be saved to filepath(file_dir/filename) if given.
    Return corpus dataframe whether the filepath is given.
    You can use this method to create corpus.parquet after load and chunk using Llama Index.

    :param langchain_documents: List of langchain documents.
    :param output_filepath: Optional filepath to save the parquet file.
        If None, the function will return the processed_data as pd.DataFrame, but do not save as parquet.
        File directory must exist. File extension must be .parquet
    :param upsert: If true, the function will overwrite the existing file if it exists.
        Default is False.
    :return: Corpus data as pd.DataFrame
    """

    corpus_df = pd.DataFrame(list(map(lambda doc: {
        'doc_id': str(uuid.uuid4()),
        'contents': doc.page_content,
        'metadata': add_essential_metadata(doc.metadata)
    }, langchain_documents)))

    if output_filepath is not None:
        save_parquet_safe(corpus_df, output_filepath, upsert=upsert)

    return corpus_df



import os
import pathlib

import click
from datasets import load_dataset


@click.command()
@click.option('--save_path', type=str, default=pathlib.PurePath(__file__).parent, help='Path to save sample hotpotqa dataset.')
def load_hotpotqa_dataset(save_path):
    # set file path
    file_path = "gnekt/hotpotqa_small_sample_autorag"

    # load dataset
    corpus_dataset = load_dataset(file_path, "corpus")['train'].to_pandas()
    qa_validation_dataset = load_dataset(file_path, "qa")['validation'].to_pandas()

    # save corpus data
    if os.path.exists(os.path.join(save_path, "corpus.parquet")) is True:
        raise ValueError("corpus.parquet already exists")
    if os.path.exists(os.path.join(save_path, "qa.parquet")) is True:
        raise ValueError("qa.parquet already exists")
    corpus_dataset.to_parquet(os.path.join(save_path, "corpus.parquet"), index=False)
    qa_validation_dataset.to_parquet(os.path.join(save_path, "qa_validation.parquet"), index=False)


if __name__ == '__main__':
    load_hotpotqa_dataset()



from typing import List, Dict, Callable

import pandas as pd

from autorag.nodes.queryexpansion.base import query_expansion_node

multi_query_expansion_prompt = """You are an AI language model assistant. 
    Your task is to generate 3 different versions of the given user 
    question to retrieve relevant documents from a vector  database. 
    By generating multiple perspectives on the user question, 
    your goal is to help the user overcome some of the limitations 
    of distance-based similarity search. Provide these alternative 
    questions separated by newlines. Original question: {question}"""


@query_expansion_node
def multi_query_expansion(queries: List[str],
                          generator_func: Callable,
                          generator_params: Dict,
                          prompt: str = multi_query_expansion_prompt) -> List[List[str]]:
    """
    Expand a list of queries using a multi-query expansion approach.
    LLM model generate 3 different versions queries for each input query.

    :param queries: List[str], queries to decompose.
    :param generator_func: Callable, generator functions.
    :param generator_params: Dict, generator parameters.
    :param prompt: str, prompt to use for multi-query expansion.
        default prompt comes from langchain MultiQueryRetriever default query prompt.
    :return: List[List[str]], list of expansion query.
    """
    full_prompts = []
    for query in queries:
        if bool(prompt):
            full_prompt = f"prompt: {prompt}\n\n question: {query}"
        else:
            full_prompt = multi_query_expansion_prompt.format(question=query)
        full_prompts.append(full_prompt)
    input_df = pd.DataFrame({"prompts": full_prompts})
    result_df = generator_func(project_dir=None, previous_result=input_df, **generator_params)
    answers = result_df['generated_texts'].tolist()
    results = list(map(lambda x: get_multi_query_expansion(x[0], x[1]), zip(queries, answers)))
    return results


def get_multi_query_expansion(query: str, answer: str) -> List[str]:
    try:
        queries = answer.split("\n")
        queries.insert(0, query)
        return queries
    except:
        return [query]



from autorag.nodes.retrieval import bm25, vectordb
from autorag.nodes.retrieval.run import run_retrieval_node
from autorag.schema import Node
from autorag.schema.module import Module
from autorag.schema.node import extract_values_from_nodes, module_type_exists


def test_get_param_combinations():
    modules = [
        Module.from_dict(
            {'module_type': 'bm25', 'key2': ['value1', 'value2'], 'key3': 'value3', 'key4': ['value4', 'value5']}),
        Module.from_dict(
            {'module_type': 'vectordb', 'key5': ['value6', 'value6', 'value7'], 'key6': 'value8'}
        )
    ]
    node = Node(node_type='retrieval', strategy={'strategy_key': 'strategy_value'}, node_params={'param1': 'value1'},
                modules=modules)

    modules, module_node_params = node.get_param_combinations()

    assert isinstance(module_node_params, list)
    assert isinstance(modules, list)
    assert len(modules) == 6
    assert modules == [bm25, bm25, bm25, bm25, vectordb, vectordb]
    bm25_solution = [{'param1': 'value1', 'key2': 'value1', 'key3': 'value3', 'key4': 'value4'},
                     {'param1': 'value1', 'key2': 'value1', 'key3': 'value3', 'key4': 'value5'},
                     {'param1': 'value1', 'key2': 'value2', 'key3': 'value3', 'key4': 'value4'},
                     {'param1': 'value1', 'key2': 'value2', 'key3': 'value3', 'key4': 'value5'}]
    vectordb_solution = [{'param1': 'value1', 'key5': 'value6', 'key6': 'value8'},
                         {'param1': 'value1', 'key5': 'value7', 'key6': 'value8'}]
    for module, module_params in zip(modules, module_node_params):
        if module == bm25:
            assert module_params in bm25_solution
            bm25_solution.pop(bm25_solution.index(module_params))
        elif module == vectordb:
            assert module_params in vectordb_solution
            vectordb_solution.pop(vectordb_solution.index(module_params))
        else:
            raise ValueError(f"Module {module} is not supposed to be here.")
    assert len(bm25_solution) == 0
    assert len(vectordb_solution) == 0


def test_from_dict():
    node_dict = {
        'node_type': 'retrieval',
        'strategy': {'strategy_key': 'strategy_value'},
        'modules': [{'module_type': 'bm25'}, {'module_type': 'bm25', 'key2': 'value2'}],
        'extra_param': 'extra_value'
    }

    node = Node.from_dict(node_dict)

    assert node.node_type == 'retrieval'
    assert node.run_node == run_retrieval_node
    assert node.strategy == {'strategy_key': 'strategy_value'}
    assert node.node_params == {'extra_param': 'extra_value'}
    assert all(isinstance(mod, Module) for mod in node.modules)
    assert len(node.modules) == 2
    assert node.modules[0].module_param == {}
    assert node.modules[1].module_param == {'key2': 'value2'}


def test_find_embedding_models():
    nodes = [
        Node.from_dict({
            'node_type': 'retrieval',
            'param1': 'value1',
            'strategy': {
                'metrics': ['retrieval_f1', 'retrieval_recall'],
            },
            'modules': [
                {
                    'module_type': 'bm25'
                },
                {
                    'module_type': 'bm25',
                    'param2': 'value2',
                    'embedding_model': ['model1', 'model2'],
                },
                {
                    'module_type': 'bm25',
                    'param2': 'value3',
                    'param3': ['value4', 'value5'],
                    'embedding_model': ['model1', 'model3'],
                }
            ]
        }),
        Node.from_dict({
            'node_type': 'retrieval',
            'strategy': {
                'metrics': ['retrieval_f1'],
            },
            'modules': [
                {
                    'module_type': 'bm25',
                    'param2': 'value2',
                    'embedding_model': ['model1', 'model3', 'model4']
                }
            ]
        })
    ]
    embedding_models = extract_values_from_nodes(nodes, 'embedding_model')
    assert set(embedding_models) == {'model1', 'model2', 'model3', 'model4'}


def test_find_llm_models():
    nodes = [
        Node.from_dict({
            'node_type': 'retrieval',
            'param1': 'value1',
            'strategy': {
                'metrics': ['retrieval_f1', 'retrieval_recall'],
            },
            'modules': [
                {
                    'module_type': 'bm25'
                },
                {
                    'module_type': 'bm25',
                    'param2': 'value2',
                    'llm': 'model1',
                },
                {
                    'module_type': 'bm25',
                    'param2': 'value3',
                    'param3': ['value4', 'value5'],
                    'llm': ['model1', 'model3'],
                }
            ]
        }),
        Node.from_dict({
            'node_type': 'retrieval',
            'strategy': {
                'metrics': ['retrieval_f1'],
            },
            'modules': [
                {
                    'module_type': 'bm25',
                    'param2': 'value2',
                    'llm': ['model1', 'model2', 'model4']
                }
            ]
        })
    ]
    llm_models = extract_values_from_nodes(nodes, 'llm')
    assert set(llm_models) == {'model1', 'model2', 'model3', 'model4'}
    assert module_type_exists(nodes, 'bm25') is True
    assert module_type_exists(nodes, 'bm26') is False
